#!/usr/bin/env ruby
#   SyntenyMapper.rb - Version 1.4.0
#   Generates a CMAP (Contig Mapping) file given a FASTA genome and target file as input.
#
#   Code Copyright (C) 2014, 2023 Daniel Veltri (dan.veltri@gmail.com)
#
#   This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  This is free software, and you are welcome to redistribute it under certain
#  conditions; for details see the LICENSE.txt file that should have been
#  distributed with this script.
#===============================================
# Created on 9.18.2014, Last Updated: 7.21.2023
# See Manual PDF for recent modifications.
#===============================================
# Basic Usage: ruby SyntenyMapper.rb <targets.fasta> <genome.fasta> <base_output_name> [options]
#
# Dependencies: BioRuby GEM, NCBI-BLAST+ (and blastall from legacy BLAST) binaries.
# *WARNING* User needs to adjust user-set variables before running script!
# See the SimpleSynteny Command Line Manual for details.
#===============================================
VERSION='1.4.0' # SyntenyMapper version for report

#==================================================
# REQUIRED LIBRARIES
require 'bio'
require 'optparse'

#==================================================
# USER-SET VARIABLES
LOCAL_MAKEBLASTDB_LOC = 'makeblastdb' # path to NCBI-BLAST+ makeblastdb
LOCAL_BLASTN_LOC = 'blastn' # path to NCBI-BLAST+ blastn
LOCAL_TBLASTN_LOC = 'tblastn' # path to NCBI-BLAST+ tblastn
CLEAN_DB = true # remove BLAST DB files when done?

#========================
# HELPER FUNCTIONS

# On Entry: 	User should set the LOCAL_MAKEBLASTDB_LOC, LOCAL_BLASTN_LOC, and LOCAL_TBLASTN_LOC paths to the proper program install locations
# On Exit: 	Will check that makeblastdb, blastn, and tblastn programs are accessible from the command line and return the program version number.
#			Warns if program numbers are not all the same.
def check_blast_programs(makeblastdb_path, blastn_path, tblastn_path)
  blast_vr = nil
  begin
    if `#{makeblastdb_path} -version` =~ /makeblastdb\:\s([0-9\.\+]+)/
      blast_vr = Regexp.last_match(1).chomp
    else
      raise "Could not run makeblastdb located at #{makeblastdb_path} - please check that the program was installed correctly and the correct path is given for LOCAL_MAKEBLASTDB_LOC variable near top of script!"
    end
    if `#{blastn_path} -version` =~ /blastn\:\s([0-9\.\+]+)/
      puts "WARNING! MAKEBLASTDB and BLASTN versions appear to be different!" if blast_vr != Regexp.last_match(1).chomp
    else
      raise "Could not run blastn located at #{LOCAL_BLASTN_LOC} - please check that the program was installed correctly and the correct path is given for LOCAL_BLASTN_LOC variable near top of script!"
    end
    if `#{tblastn_path} -version` =~ /tblastn\:\s([0-9\.\+]+)/
      puts "WARNING! MAKEBLASTDB and TBLASTN versions appear to be different!" if blast_vr != Regexp.last_match(1).chomp
    else
      raise "Could not run tblastn at #{tblastn_path} - please check that the program was installed correctly and the correct path is given for LOCAL_TBLASTN_LOC variable near top of script!"
    end
    if blast_vr == nil
		raise "Was unable to set BLAST+ program version - please verify programs are installed correctly and proper paths are provided to the script near the top!"
    else
	  return blast_vr
    end
  rescue Exception => err
    puts "ERROR testing NCBI-BLAST+ programs: #{err.message}\n"
	exit(1)
  end
end

# On Entry: Defaults for evalue_cutoff, nthreads, and query_coverage_cutoff provided
# On Exit: Will exit to console after printing program help message to screen
def console_message(evalue_cutoff=0.001 , nthreads=1, query_coverage_cutoff=30.0)
  puts "\nSyntenyMapper Usage:"
  puts "ruby SyntenyMapper.rb -q <query_genes_fasta> -g <subject_genome/contigs_fasta> -p <output_prefix_name> [optional flags]"
  puts "\nRequired Arguments:"
  puts "-q|--query <query_genes_fasta_file> :\t\tFASTA file with gene nucleotide sequences (if proteins use the -p flag) to search for."
  puts "-g|--genome <subject_genome/contigs_fasta_file> :\tBlast Nucleotide Database for organism of interest. Create with makeblastdb."
  puts "-p|--prefix <output_prefix_name> :\t\t\tPrefix to name file output."
  puts "\nCommon Option Flags:"
  puts "-e|--evalue <num/float> :\tAdjust the E-value cutoff for BLAST (default: #{evalue_cutoff})."
  puts "-c|--cov <num/float> :\t\tAdjust the Minimum Query Coverage / Percent Identity for a valid BLAST hit to be shown (default: #{query_coverage_cutoff}, set to 0 to turn off)."
  puts "-f|--flip :\t\t\tShow both forward and flipped (reverse compliment) mappings when printing (default: off)."
  puts "--nthreads <num> :\t\tPass number of threads to use for BLAST (Default: #{nthreads})."
  puts "--ungap :\t\t\tHave BLAST perform gapped alignments? True/False (default: T)."
  puts "--gene <string> :\t\tFilter to show ONLY results with gene name matching '<string>' as gene name (default: off)."
  puts "--best_hsp :\t\t\tPrint top scoring HSP only, cancel expansion to other valid hits in same strand direction of best HSP (default: off)"
  puts "--no_coord :\t\t\tDon't print out the full coordinates, just preview contigs and gene names (default: off).\n\t\t\t\tCan't use with --flip option."
  puts "-h|--help :\t\t\tPrint this message."
  puts "\n\nAdvanced Option Flags:"
  puts "-t|--tol <num/float> :\t\tTolerance for what percent of query target length an HSP must cover to be counted as close enough to an 'exact match'.\n\t\t\t\t(MUST range between 50.0 and 100.0%; default: 95.0)."
  puts "--prot :\t\t\tForce Protein Mode and use 'tblastn' if <query_genes_fasta> contains only protein sequences and isn't auto-detected.\n\t\t\t\tMultiplies coordinates x3 to convert from AA to NA and assumes no exon splicing (default: off).\n\t\t\t\tCan't use with --prot_noa_djust option."
  puts "--prot_no_adjust :\t\tUse 'tblastn' but DO NOT multiply coordinates x3 to convert from AA to NA (default: off).\n\t\t\t\tCan't use with --prot option."
  puts "-d|--debug :\t\t\tPrint verbose program output information (default: off)."
  puts "\n\n**This program comes with no warranty**"
  puts "Please see the attached documentation for CMAP file format details."
  puts "Contact Dan Veltri (dan.veltri@gmail.com) if you have problems with the program.\n\n"
end

# On Entry: cmap_line is a valid CMAP string
# On Exit: cmap_line is flipped including coordinates and gene directions and returned as string
def flip_contig(cmap_line)
  cstart = 0
  cend = 0
  genes = []
  if cmap_line =~ /(\{)(\d+)(.*?)(\d+)(\})/
    cstart = $2.to_i
    cend = $4.to_i
    genes = $3.split(" ]  [ ")
  end

  pstring = "{#{cstart}  "
  revgenes = genes.reverse
  for i in 0 ... genes.length
    e = revgenes[i]
    if e =~ /(\d+)(\|)(\d+)(.*\).*?)(\d+)(\|)(\d+)/
      gfstart = $1.to_i
      gfend = $7.to_i
      gstart = $3.to_i
      gend = $5.to_i
      gstring = $4
      raise "\n\nWARNING! Flipped gene coords don't match contig length!\n#{e}\nContig Len: #{cend}\nGene Start: #{gfstart} End: #{gfend}\n\n\n" if (cend - gfend + 1) <= 0
      if gstring =~ /(\(\d+\|\d+)(.*)(\s+==>\s+)(\d+\|\d+\))/
        pstring += "[ #{(cend - gfend + 1)}|#{(cend - gend + 1)} #{$1}  <== #{$2.strip}  #{$4} #{(cend - gstart + 1)}|#{(cend - gfstart + 1)} ]  "
      elsif gstring =~ /(\(\d+\|\d+)(\s+<==\s+)(.*\s+)(\d+\|\d+\))/
        pstring += "[ #{(cend - gfend + 1)}|#{(cend - gend + 1)} #{$1}  #{$3.strip} ==>  #{$4} #{(cend - gstart + 1)}|#{(cend - gfstart + 1)} ]  "
      else
        raise "ERROR in flip_contig! Invalid gene format: #{gstring}"
      end
    else
      raise "ERROR in flip_contig! Invalid contig format: #{e}"
    end
  end
  pstring += "#{cend}}\n"
  return pstring
end


# On Entry: fastafile is file name of valid fasta file, target_definition is definition line to retrieve from fastafile and MUST be there else function will raise error!
# On Exit: block is returned with target_definition's sequence, target definitions reverse complement and the sequence length
def get_contig(fastafile, target_definition)
  count = 0
  Bio::FastaFormat.open(fastafile).each do |f|
    count += 1
    return f.seq, f.naseq.reverse_complement.upcase, f.seq.length if f.definition == target_definition
  end

  # If we reach here it's an error...
  if count == 0
    raise "ERROR in get_contig_length! Empty target fasta file given!"
  else
    raise "ERROR in get_contig_length! Invalid definition '#{target_definition}' given!"
  end
end


# Splits hits into uniq ranges to consider in order to handle target repeats per reviewer request
# Definitely should code a more efficient way to do this given time.
# On entry:  hit from BioRuby BLAST report is provided for one gene on one contig
#                   query_len is the full length of the query target
#                   evalue_cutoff is threshold under which hits must be to be considered
#                   full_match_tolerance_pct is the percent of query_len that determines if an HSP is an "exact match" and gets its own range (default=95 or 95% of query_len, 50-100% allowed)
# On exit: array container contains arrays with start-end ranges for the program to consider as a uniq hit region e.g. [ [bp_start, bp_end], ... , [bp_start, bp_end] ]
#
# Basic approach:
# 1) We gather locations and ranges of hits and separate these based on direction.
# 2) We consider overlaping regions of hits in different directions, and chose the one containing an HSP of lowest e-value
# 3) We merge overlapping hits and remove duplicates
def segment_hits(hit, query_len, evalue_cutoff = 0.001, full_match_tolerance_pct=95.0)
  if full_match_tolerance_pct < 50 || full_match_tolerance_pct > 100
    raise "Error in segment_hits: invalid 'full_match_tolerance_pct' provided (#{full_match_tolerance_pct})! Must be a value between 50-100%."
  end

  tolerance = full_match_tolerance_pct / 100.0
  range_info = Struct.new(:evalue, :gene_start, :gene_end, :seg_start, :seg_end, :dir) #holds hit results
  fwd_ranges, rev_ranges = Array.new, Array.new # collection of  direction_specific range_info structs
  fwd_section, rev_section = Array.new, Array.new # holds ranges generated from hit data each as an array [bp_start, bp_end]
  fwd_evalue, rev_evalue = Array.new, Array.new # corresponding evalues to fwd_section elements
  merged_section = Array.new # holds final merged fwd and rev ranges for both directions (when present)

  hit.each do |hsp|
    unless hsp.hit_frame < 0
      fwd_ranges << range_info.new(hsp.evalue, hsp.query_from, hsp.query_to, hsp.hit_from, hsp.hit_to, '==>') if hsp.evalue <= evalue_cutoff
    else
      rev_ranges << range_info.new(hsp.evalue, hsp.query_from, hsp.query_to, hsp.hit_from, hsp.hit_to, '<==') if hsp.evalue <= evalue_cutoff
    end
  end

  # Get forward ranges
  this_min, contig_seg_end, ev = 0, 0, 0
  contig_seg_start = 999999999999
  unless fwd_ranges.empty?
    fwd_ranges.each do |r|
      ev = r.evalue if r.evalue < ev
      if (query_len - (r.gene_end - r.gene_start).abs <= (tolerance * query_len)) # is the hsp a full copy of the query?
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        fwd_section << [contig_seg_start, contig_seg_end]
        fwd_evalue << ev
        this_min = 0
        ev = 999999999999
      elsif r.gene_end > this_min
        contig_seg_start = r.seg_start if r.seg_start < contig_seg_start
        contig_seg_end = r.seg_end if r.seg_end > contig_seg_end
      else
        fwd_section << [contig_seg_start, contig_seg_end]
        fwd_evalue << ev
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        this_min = 0
        ev = 999999999999
      end
      this_min = r.gene_end
    end
    fwd_section << [contig_seg_start, contig_seg_end]
    fwd_evalue << ev
  end

  # Get reverse ranges
  this_min, contig_seg_end, ev = 999999999999, 0, 999999999999
  contig_seg_start = 999999999999
  unless rev_ranges.empty?
    rev_ranges.each do |r|
      ev = r.evalue if r.evalue < ev
      if (query_len - (r.gene_start - r.gene_end).abs <= (tolerance * query_len)) # is the hsp a full copy of the query?
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        fwd_section << [contig_seg_start, contig_seg_end]
        fwd_evalue << ev
        this_min = 999999999999
        ev = 999999999999
      elsif r.gene_end < this_min
        contig_seg_start = r.seg_start if r.seg_start < contig_seg_start
        contig_seg_end = r.seg_end if r.seg_end > contig_seg_end
      else
        rev_section << [contig_seg_start, contig_seg_end]
        rev_evalue << ev
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        this_min = 999999999999
        ev = 999999999999
      end
      this_min = r.gene_end
    end
    rev_section << [contig_seg_start, contig_seg_end]
    rev_evalue << ev
  end


  # Save regions in both directions - if both occur in the same range we choose the one with the lower E-value
  if fwd_section.length == 0
    merged_section = rev_section.uniq
  elsif rev_section.length == 0
    merged_section = fwd_section.uniq
  else # we check for overlapping sections and take one with lower e-value
    fwd_section.uniq.each_index do |i|
      rev_section.uniq.each_index do |j|
        if( [ fwd_section[i][0], rev_section[j][0] ].max <= [ fwd_section[i][1], rev_section[j][1] ].min )
          if fwd_evalue[i] <= rev_evalue[j]
            merged_section << [fwd_section[i][0], fwd_section[i][1]]
          else
            merged_section << [rev_section[j][0], rev_section[j][1]]
          end
        else # both segments don't overlap
          merged_section << [fwd_section[i][0], fwd_section[i][1]]
          merged_section << [rev_section[j][0], rev_section[j][1]]
        end
      end
    end
  end

  # Merge overlapping regions
  merged = true
  merged_section.uniq!
  while(merged && merged_section.length > 1) # continue to loop and check as long as a segment is merged
    merged = false
    for i in 0 ... (merged_section.length - 1)
      for j in (i + 1) ... merged_section.length
        if( [ merged_section[i][0], merged_section[j][0] ].max <= [ merged_section[i][1], merged_section[j][1] ].min )
          merged = true
          low = [merged_section[i][0], merged_section[j][0]].min
          high = [merged_section[i][1], merged_section[j][1]].max
          merged_section[i][0] = low
          merged_section[j][0] = low
          merged_section[i][1] = high
          merged_section[j][1] = high
          break
        end
      end
    end
    merged_section.uniq!
  end
  return merged_section
end

#========================
# MAIN

# General containers
query_filename = nil
subj_filename = nil
output_prefix = nil
all_hits = []
nhits = 0 # hits meeting evalue cutoff
idenhits = 0 # hits meeting query coverage cutoff
match_gene_name = nil # holds regex of name to match
cfinder_seqs = Hash.new # FOR SERVER ONLY
seen_gene_names = Hash.new
print_gene_counts = Hash.new
this_gene_name = ''
hit_info = Struct.new(:gene_name, :contig_num, :full_gene_start, :full_gene_end, :seg_start, :seg_end, :contig_start, :contig_end, :contig_len, :direction) #holds hit results
time_id = Time.now.to_i # used to name temp BLAST DB files

# Blast User Defaults
evalue_cutoff = 0.001
gap_align = '' # for blast+ a blank here is a default of gapped=TRUE
nthreads = 1
expand_hits = true # try to go out from best HSP and expand hit range for other valid hits in same strand direction
query_coverage_cutoff = 30.0 # min query coverage cutoff
full_match_tolerance_pct = 95.0 # percentage of query target length an HSP can cover and be counted as close enough to an "exact match" (will get its own range and be displayed)
prot_search = false #force tblastn

# Other User Defaults
filter_gene_name = false # set mode to only print results for a given gene name?
coords = true # return coordinates?
adjust_aa2na = false #if the -p flag is set, multiply query coords x 3 to convert back to NA numbering. This assumes no exon splicing is present!
mul_type = 1 # set to 1 for nucleotide genes, will be set to 3 if protien query flag is thrown
flip = false # also print revcomp flipped hit cmap lines when true
silent = true #suppress debug output
cfinder_mode = false # FOR SERVER ONLY - prints CFinderSeqs.fa with seqs with hits if -cfinder flag is provided

# Print user message and get user arguments - there are some additional options used for debugging hidden by default here
puts "\n================================================================================================"
puts "SyntenyMapper Vr.#{VERSION} - Takes FASTA query and genome target files along with prefix to name file outputs and\nprints/saves valid CMAP lines for use with SyntenyDrawer."
puts "\nPlease Cite: Veltri, D., Malapi-Wight, M. and Crouch J.A. SimpleSynteny: a web-based tool for"
puts "visualization of microsynteny across multiple species. Nucleic Acids Research, 44(W1) 2016."
puts "------------------------------------------------------------------------------------------------"
blast_vr = check_blast_programs(LOCAL_MAKEBLASTDB_LOC, LOCAL_BLASTN_LOC, LOCAL_TBLASTN_LOC) # check NCBI-BLAST+ programs are available
if ARGV.length < 3 || (['-q','--query'] & ARGV).length != 1 || (['-g','--genome'] & ARGV).length != 1 || (['-p','--prefix'] & ARGV).length != 1
  console_message(evalue_cutoff, nthreads, query_coverage_cutoff)
  puts "\nMISSING REQUIRED PARAMETERS! You must provide one query FASTA (-q|--query), one genome FASTA (-g|--genome), and one output prefix (-p|--prefix) parameter!\n\n"
  exit(1)
else
  begin
    OptionParser.new do |opts|
	  opts.on('-q','--query VALUE', "FASTA file of query genes/proteins for searching") do |value|
        if value !~ /\.f[apn]s?t?a?$/i
          raise "Please ensure your query filename ense with .fasta, .fa, .fn, or .fp file extension and is in the proper format! It should contain only nucleotide or protein sequences."
		else
          query_filename = value
		end
      end
      opts.on('-g','--genome VALUE', "FASTA file of subject genome chromosomes/contigs for searching") do |value|
        if value !~ /\.f[an]s?t?a?$/i
          raise "Please ensure your genome/contig filename ense with .fasta, .fa, or .fn file extension and is in the proper format! It should contain only nucleotide sequences."
		else
          subj_filename = value
		end
      end
      opts.on('-p','--prefix VALUE', "Prefix for output filenames") do |value| output_prefix = value.gsub(/\s+/,'_') end
      opts.on('--no_coord', "Don't print out the full coordinates, just preview contigs and gene names (default: off).\n\t\t\tCan't use with --flip option") do
        if flip == true
          raise "--no_coord and --flip arguments may not be used together!"
        else
          coords = false
	    end
      end
      opts.on('--prot', "Force Protein Mode and use 'tblastn' if <my_query_genes_fasta> contains protein sequences.\n\t\t\tMultiplies coordinates x3 to convert from AA to NA and assumes no exon splicing (default: off).\n\t\t\tCan't use with -p_noadjust option.") do
        prot_search = true
        mul_type = 3
        adjust_aa2na = true
	  end
      opts.on('--prot_no_adjust', "Use 'tblastn' but DO NOT multiply coordinates x3 to convert from AA to NA (default: off).\n\t\t\tCan't use with -p option.") do
        if prot_search == true
          raise "--prot and --prot_no_adjust arguments may not be used together!"
        else
          prot_search = true
        end
      end
      opts.on('-b', '--best-hsp', "Print top HSP only, cancel expansion to other valid hits in same strand direction of best HSP (default: off)") do expand_hits = false end
      opts.on('-e', '--evalue VALUE', Float, "Set the E-value cutoff for BLAST (default: #{evalue_cutoff})") do |value|
        if value.to_f < 0.0 || value.to_f > 100.0
          raise "Evalue cutoff must be between 0.0 and 100.0! Received: -evalue #{value.to_f}"
        else
          evalue_cutoff = value.to_f
		end
	  end
      opts.on('-c', '--cov VALUE', Float, "Adjust the Minimum Query Coverage / Percent Identity for a valid BLAST hit to be shown (default: #{query_coverage_cutoff}, set to 0 to turn off).") do |value|
        if value.to_f < 0.0 || value.to_f > 100.0
          raise "Coverage cutoff must be between 0.0 and 100.0! Received: -cov #{value.to_f}"
        else
          query_coverage_cutoff = value.to_f
        end
	  end
      opts.on('-f', '--flip', "Show reverse compliment 'flipped' hit results as well (default: off)") do flip = true end
      opts.on('--gene VALUE', String, "Show ONLY query results that match '<string>' as gene name (default: show all)") do |value| match_gene_name = Regexp.new(value, true) end
      opts.on('--ungap', "Have BLAST perform an ungapped alignment (default: gapped)") do gap_align = '-ungapped' end
      opts.on('-n', '--nthreads VALUE', Integer, "Set number of CPU threads to use with BLAST (Default: #{nthreads})") do |value|
        if value.to_i < 1
          raise "nthreads must be >= 1! Received: -nthreads #{value.to_i}"
		else
          nthreads = value.to_i
        end
      end
	  opts.on('-t', '--tol VALUE', Float, "Tolerance for what percent of query target length an HSP must cover to be counted as close enough to an 'exact match' (MUST range between 50.0 and 100.0%; default: 95.0)") do |value|
        if value.to_f < 50.0 || value.to_f > 100.0
          raise "Value for tolerance must be between 50.0 and 100.0! Received: -tol #{value.to_f}"
		else
          full_match_tolerance_pct = value.to_f
        end
      end
	  opts.on('-d', '--debug', "Print extra information (default: off)") do silent = false end
      opts.on("--cfinder", "Contig Finder Mode (server use only!; default: off)") do cfinder_mode = true end
      opts.on('-h', '--help', "Prints this help message") do console_message(evalue_cutoff, nthreads, query_coverage_cutoff); exit end
    end.parse!(ARGV)
  rescue Exception => err
  	console_message(evalue_cutoff, nthreads, query_coverage_cutoff)
    puts "ERROR: #{err.message}\n"
	exit(1)
  end
end
if ARGV.length > 0
  console_message()
  puts "\n\nERROR: The following unknown parameter(s) were submitted:\n#{ARGV.join("\n")}\n\nPlease correct and try again!"
  exit(1)
end

puts "\nWARNING: Only using best HSP for each hit!\n\n" unless silent == true || expand_hits == true

#Check if queries are NA or Protein also see if we have duplicate gene names
query_ff = Bio::FlatFile.auto(query_filename)
query_ff.each do |s|
  if seen_gene_names.include?(s.definition) then seen_gene_names[s.definition] += 1 else seen_gene_names[s.definition] = 1; print_gene_counts[s.definition] = 0 end
  if prot_search == false && s.seq =~ /[EFILMPQZ]/i then prot_search = true end
end
if silent == false
  if prot_search == false
    puts "Detected a nucleotide query file!"
    print "Blasting (w/#{nthreads}CPU) genes (Query coverage cutoff #{query_coverage_cutoff}; E-value cutoff #{evalue_cutoff}; Gapped Aln: #{gap_align}) in #{query_filename} to #{subj_filename}...\n"
  else
    puts "Detected a protein query file!"
    print "[T]Blasting (w/#{nthreads}CPU) proteins (Quercy coverage cutoff #{query_coverage_cutoff}; E-value cutoff #{evalue_cutoff}; Gapped Aln: #{gap_align}) in #{query_filename} to #{subj_filename}...\n"
  end
end

# Report files for output
blast_report_out = File.open(output_prefix + "_with_query_#{query_filename.gsub(/^.*\//,'').gsub(/\.f[astq]+/i,'')}_BLAST_REPORT.csv","w")
skipped_report_out = File.open(output_prefix + "_skipped_genes.txt","w") #File to record any genes skipped due to low sequence identity
mask_report_out = File.open(output_prefix + "_with_query_#{query_filename.gsub(/^.*\//,'').gsub(/\.f[astq]+/i,'')}_MASKS.txt","w") # File to save BLAST output for user-reporting
blast_report_out.write("qseq_name,sseq_name,alignment_length,num_gaps,qstart,qend,sstart,send,evalue,bitscore\n") #report header

# Generate BLAST DB
begin
  `#{LOCAL_MAKEBLASTDB_LOC} -dbtype nucl -in #{subj_filename} -out #{output_prefix}_#{time_id}_db` unless File.file?("#{time_id}.nin")
  puts "Finished making DB! #{time_id}" unless silent == true
rescue
  raise "ERROR! MAKEBLASTDB failed to complete query!\n Is makeblastdb properly installed at: #{LOCAL_MAKEBLASTDB_LOC} and is your GENOME query '#{subj_filename}' a valid nucleotide FASTA/Q file?"
end

# Perform BLAST+ query on local system and read reports into bioruby blast factory
if prot_search == true
  begin
    `#{LOCAL_TBLASTN_LOC} -query #{query_filename} -db #{output_prefix}_#{time_id}_db -out #{output_prefix}_#{time_id}_tblastn_output.txt -num_threads #{nthreads} -outfmt 5 #{gap_align} -evalue #{evalue_cutoff}`
	blast_report_out.write("TBLASTN Vr. #{blast_vr} RESULTS\n")
  rescue
    raise "ERROR! TBLASTN failed to complete search!\n Is tblastn properly installed at: #{LOCAL_TBLASTN_LOC} and is your QUERY file '#{query_filename}' a valid protein FASTA file?"
  end
else
  begin
    `#{LOCAL_BLASTN_LOC} -query #{query_filename} -db #{output_prefix}_#{time_id}_db -out #{output_prefix}_#{time_id}_blastn_output.txt -num_threads #{nthreads} -outfmt 5 #{gap_align} -evalue #{evalue_cutoff}`
	blast_report_out.write("BLASTN Vr. #{blast_vr} RESULTS\n")
  rescue
    raise "ERROR! BLASTN failed to complete search!\n Is blastn properly installed at: #{LOCAL_BLASTN_LOC} and is your QUERY file '#{query_filename}' a valid nucleotide FASTA file?"
  end
end

# Read in results to bioruby blast factory report
begin
  if prot_search == true
    blast_reports = Bio::Blast::Report.new(File.read("#{output_prefix}_#{time_id}_tblastn_output.txt"))
  else
	blast_reports = Bio::Blast::Report.new(File.read("#{output_prefix}_#{time_id}_blastn_output.txt"))
  end
rescue
  raise "ERROR! Was not able to read in the blast search results! Please check if #{time_id}_output.txt was written and verify BLAST programs are working correctly."
end

# Cycle through queries
query_ff.rewind
query_ff.each do |f|

  next if f.definition.to_s !~ match_gene_name && filter_gene_name == true

  mask_report_out.write("GENE: #{f.definition} - LENGTH: #{f.seq.length}\nSEQ  #{f.seq}\n")
  this_report = []
  blast_reports.each_hit {|h| this_report << h if h.query_def == f.definition }
  query_len = f.seq.length
  best_identity = 0.0
  best_iden_contig = ''
  best_bitscore = 0.0
  valid_hsp_evalue = false
  any_hits = false

  # Adjust gene name and see if we've seen it before (append .1, .2 etc. for repeats)
  if seen_gene_names[f.definition] > 1
    print_gene_counts[f.definition] += 1
    this_gene_name = f.definition.gsub(/\s+/,'_') + ".#{print_gene_counts[f.definition]}"
  else
    print_gene_counts[f.definition] = 1
    this_gene_name = f.definition.gsub(/\s+/,'_')
  end

  puts "\n-------------------------\n#{this_gene_name}" unless silent == true
  if prot_search == true
    puts "Query Length: #{f.seq.length}aa" unless silent == true
  else
    puts "Query Length: #{f.seq.length}bp" unless silent == true
  end

  this_report.each do |hit|
    hit_def = hit.target_def
    mask_report_out.write("CONTIG: #{hit_def}\n")
    puts "\nCONTIG: #{hit_def}\n" unless silent == true

    hit_search_segments = segment_hits(hit, query_len, evalue_cutoff, full_match_tolerance_pct) # contains 2d array of ranges to search, [  [bp_start, bp_end] , [bp_start, bp_end] , ... ]

    hit_search_segments.each_index do |sidx|
      dirset = false
      firsthsp_only = false
      hitdir = ''
      evals = Array.new
      lowest_subj = 0
      highest_subj = 0
      lowest_query = 0
      highest_query = 0
      fullquery_iden = 0.0

      # Figure identity of all tiled HSP's across full-length query and locate anchor segment (best scoring HSP)
      query_bit_vector = "0" * query_len #bit-vector same length of query to keep track of tiles
      mask_report_out.write("INIT #{query_bit_vector}\n")

      hit.each do |hsp|
        blast_report_out.write("#{f.definition},#{hit.target_def},#{hsp.align_len},#{hsp.gaps},#{hsp.query_from},#{hsp.query_to},#{hsp.hit_from},#{hsp.hit_to},#{hsp.evalue},#{hsp.bit_score}\n") #record hit for report

        if hsp.evalue <= evalue_cutoff && firsthsp_only == false && ( [hsp.hit_from, hit_search_segments[sidx][0]].max <= [hsp.hit_to, hit_search_segments[sidx][1]].min )

          # Lock anchor and hit direction to best/initial HSP
          if dirset == false
            best_bitscore = hsp.bit_score

            lowest_query = [hsp.query_from,hsp.query_to].min
            highest_query = [hsp.query_from,hsp.query_to].max
            lowest_subj = [hsp.hit_from,hsp.hit_to].min
            highest_subj = [hsp.hit_from,hsp.hit_to].max

            unless hsp.hit_frame < 0
              hitdir = '==>'
            else
              hitdir = '<=='
            end
            puts "Gene Direction of Best HSP: #{hitdir}" unless silent == true
            puts "Best HSP Ranges - Query: [#{lowest_query},#{highest_query}] Subj: [#{lowest_subj},#{highest_subj}]" unless silent == true
            valid_hsp_evalue = true #signal at least one valid hsp found
            dirset = true
          end


          # Only consider hits in same direction as initial/best HSP
          if (hsp.hit_frame < 0 && hitdir == '<==') || (hsp.hit_frame >= 0 && hitdir == '==>')

            this_lowest_query = [hsp.query_from,hsp.query_to].min
            this_highest_query = [hsp.query_from,hsp.query_to].max
            this_lowest_subj = [hsp.hit_from,hsp.hit_to].min
            this_highest_subj = [hsp.hit_from,hsp.hit_to].max
            puts "This HSP - Query: [#{this_lowest_query},#{this_highest_query}] Subj: [#{this_lowest_subj},#{this_highest_subj}]" unless silent == true

            # Check for valid match
            pass_check = false
            if ((this_lowest_query <= lowest_query && this_lowest_subj <= lowest_subj && this_highest_subj <= highest_subj && this_lowest_subj >= (lowest_subj - 2 * query_len)) || (this_highest_query >= highest_query && this_highest_subj >= highest_subj && this_lowest_subj >= lowest_subj && this_highest_subj <= (highest_subj + 2 * query_len))) && (hitdir == '==>')
              pass_check = true
            elsif ((this_lowest_query <= lowest_query && this_lowest_subj >= lowest_subj && this_highest_subj >= highest_subj && this_lowest_subj <= (lowest_subj + 2 * query_len)) || (this_highest_query >= highest_query && this_highest_subj <= highest_subj && this_lowest_subj <= lowest_subj && this_highest_subj >= (highest_subj - 2 * query_len))) && (hitdir == '<==')
              pass_check = true
            end
            if(pass_check == true)
              lowest_query = [this_lowest_query,lowest_query].min
              highest_query = [this_highest_query,highest_query].max
              lowest_subj = [this_lowest_subj,lowest_subj].min
              highest_subj = [this_highest_subj,highest_subj].max

              evals << hsp.evalue
              full_mask = "0" * query_len #initialize full-size mask for this HSP to query length

              # Create mask for just HSP segment, skip over query gaps, match=1, mismatch=0
              qry_mask = ''
              addgap = 0
              puts "HSP: #{hsp.qseq.gsub(/\-/,'')}" unless silent == true
              for i in 0 ... hsp.qseq.length

                if hsp.qseq[i,1] == '-'
                  addgap += 1
                elsif hsp.qseq[i,1] == 'X'
                  qry_mask += '0'
                else
                  qry_mask += '1'
                end
              end

              if hitdir == '<==' && prot_search == false
                cleaned_hsp_qseq = Bio::Sequence.auto(hsp.qseq.gsub(/\-/,'')).reverse_complement
              else
                cleaned_hsp_qseq = hsp.qseq.gsub(/\-/,'')
              end

              hsp_mask = " " * query_len
              hsp_mask[(this_lowest_query - 1),  cleaned_hsp_qseq.length] = cleaned_hsp_qseq

              full_mask[(this_lowest_query - 1), cleaned_hsp_qseq.length] = qry_mask #splice HSP segment mask into full-length mask
              query_bit_vector = (query_bit_vector.to_i(2) | full_mask.to_i(2)).to_s(2).rjust(query_len, "0") #bitwise OR with current query_bit_vector state
              mask_report_out.write("\nMASK #{full_mask}\nHSP  #{hsp_mask} - LENGTH:#{qry_mask.length}, EVALUE:#{hsp.evalue}\n")
            end
          end

          firsthsp_only = true if expand_hits == false #if we're not expanding from first HSP, stop coverage calculation now
        end
      end

      nhits += 1 if valid_hsp_evalue == true

      puts "Range of All Valid HSPs - Query: [#{lowest_query},#{highest_query}] Subj: [#{lowest_subj},#{highest_subj}]" unless silent == true

      fullquery_iden = (query_bit_vector.count('1') / query_len.to_f) * 100.0
      puts "Query Coverage/Identity: #{fullquery_iden}% (ignores X's and gaps)" unless silent == true
      mask_report_out.write("\nFNL  #{query_bit_vector} \nIDENTITY: #{fullquery_iden}\n\n")

      if best_identity < fullquery_iden
        best_identity = fullquery_iden
        best_iden_contig = hit_def
      end

      if fullquery_iden >= query_coverage_cutoff && valid_hsp_evalue == true
        any_hits = true
        idenhits += 1
        this_contig_seq, this_contig_seq_revcomp, this_contig_len = get_contig(subj_filename, hit_def)

        if adjust_aa2na == true
          diff_high = (query_len * 3) - ((query_len - highest_query) * 3)
          diff_low = ((lowest_query - 1) * 3) + 1
          all_hits << hit_info.new(this_gene_name, hit_def, 1, (query_len * 3), diff_low, diff_high, lowest_subj, highest_subj, this_contig_len, hitdir)
        else
          all_hits << hit_info.new(this_gene_name, hit_def, 1, query_len, lowest_query, highest_query, lowest_subj, highest_subj, this_contig_len, hitdir)
        end

        # SERVER ONLY - CFinder mode
        if(cfinder_mode == true && cfinder_seqs.has_key?(hit_def) == false)
          cfinder_seqs[hit_def] = this_contig_seq
        end

      end
    end
  end

  # Record if we didn't get any matches for this query
  if any_hits == false
    if best_identity < query_coverage_cutoff && valid_hsp_evalue == true#there was an HSP but didn't meet identity cutoff
      skipped_report_out.write("#{this_gene_name}: No matches satisfying full-length sequence coverage cutoff: #{query_coverage_cutoff}% (highest found: #{"%0.2f" % best_identity}%)\n")
      puts "\tHit(s) on contigs missed sequence coverage cutoff: #{"%0.3f" % best_identity} (Cutoff:#{query_coverage_cutoff})" unless silent == true

    else #gene didn't map due to no significant HSP's at this e-value
      skipped_report_out.write("#{this_gene_name}: No significant matches found at E-value cutoff: #{evalue_cutoff}\n")
      puts "\tHit(s) on all contigs did not satisfy sequence e-value cutoff: #{"%0.3f" % evalue_cutoff}" unless silent == true
    end
  end

end

# SERVER ONLY - Save contig hits if CFinder mode is on
if cfinder_mode == true
  puts "\nWriting contigs with hits out to #{output_prefix}_CFinder_Hits.fa for CFinder..." unless silent == true
  puts cfinder_seqs unless silent == true
  cfinder_out = File.open(output_prefix + "_CFinder_Hits.fa","w")
  cfinder_seqs.each {|k,v| cfinder_out.write(">#{k}\n#{v}\n") }
  cfinder_out.close
end

# Close reports
skipped_report_out.close
blast_report_out.close
mask_report_out.close

# Prep CMAP output files
fwd_out = File.open(output_prefix + "_fwd.cmap","w")
rev_out = File.open(output_prefix + "_rev.cmap","w") if flip == true

# Make sure we got a hit based on e-value and identity
if idenhits == 0
  puts "\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%"
  fwd_out.write("\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%")
  rev_out.write("\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%") if flip == true
  fwd_out.close()
  rev_out.close() if flip == true
  exit
elsif nhits == 0
  puts "\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}"
  fwd_out.write("\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}")
  rev_out.write("\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}") if flip == true
  fwd_out.close()
  rev_out.close() if flip == true
  exit
else
  print "\n DONE! (#{nhits} hits found)\n\nGenerating CMAP line(s):\n\n" unless silent == true
end

# Print CMAPs of hits in order of original genome file
index_order = []
contigs = all_hits.uniq.collect {|h| h.contig_num }
Bio::FlatFile.auto(subj_filename).each {|f| res = contigs.find_index(f.definition); index_order << res unless res == nil}
index_order.each do |idx|
  contig = contigs[idx]
  cgenes = all_hits.select{|c| c.contig_num == contig}
  cgenes.sort! {|a,b| a.contig_start <=> b.contig_start}
  last_name = ''
  print_string = '{1 '
  cgenes.each do |g|

    if g.direction == '<=='
      print_gene_format = "#{g.direction} #{g.gene_name}"
    elsif g.direction == '==>'
      print_gene_format = "#{g.gene_name} #{g.direction}"
    else
      raise "ERROR IN DIRECTION: #{g}"
      exit(1)
    end

    if coords == true
        print_string += " [ #{[(g.contig_start - (g.seg_start - g.full_gene_start)),1].max}|#{g.contig_start} (#{([(g.contig_start - g.seg_start - 1),1].min).abs}|#{g.seg_start}  #{print_gene_format}  #{g.seg_end}|#{g.full_gene_end}) #{g.contig_end}|#{[(g.contig_end + (g.full_gene_end - g.seg_end)),g.contig_len].min} ] "
    else
      print_string += "[ #{print_gene_format} ]" unless g.gene_name == last_name
      last_name = g.gene_name
    end
  end

  print_string += " #{cgenes[0].contig_len}}\n"

  if flip == true
    puts "Fwd: #{contig} " + print_string
    fwd_out.write("#{contig} #{print_string}")
    puts "\nRev: #{contig} " + flip_contig(print_string) + "\n\n"
    rev_out.write("#{contig} " + flip_contig(print_string))
  else
    puts "#{contig} " + print_string
    fwd_out.write("#{contig} #{print_string}")
  end
end
fwd_out.close
rev_out.close if flip == true

# Cleanup BLAST DB Files
Dir.glob("#{output_prefix}_#{time_id}*").each {|f| File.delete(f)} if CLEAN_DB == true

#========================
# END PROGRAM
#========================